# -*- coding: utf-8 -*-
import logging
from datetime import date

import formencode

from pylons import app_globals as g

from mpulsweb.lib.validators import BundleActionForm
from mpulsweb.lib.base import c, render, request, session, BaseController
from mpulsweb.lib.db import db
from mpulsweb.lib.security import checkRole
from mpulsweb.lib.helpers import url
from mpulsweb.lib.translation import _, ungettext
from mpulsweb.model.user import UserListObject
from mpulsweb.model.agency import MpulsAgencyOverview
from mpulsweb.model.phase import PhaseFactory


log = logging.getLogger(__name__)

def get_sort(params):
    '''Return sort field and sort order. First try to return a userdefined sort
    and order. If user does not provide one of there params, try to return the
    safed params in session. If session is also empty return default params.
    '''
    session_options = session.get('AGENCY_OVERVIEW_SEARCHOPTIONS', {})
    sort = params.get('sort_field', session_options.get('sort_field'))
    order = params.get('sort_order', session_options.get('sort_order'))
    session_options['sort_field'] = sort
    session_options['sort_order'] = order
    session.save()
    return sort, order

def default_phases():
    phases = []
    for pp in g.mpuls_config.get('phases', 'phases'):
        pairs = g.mpuls_config.get('phases', 'pairs')
        for p in pairs[0].get(str(pp), [-1]):
            phases.append(p)
    phases.append(-1)
    return phases

def parseSearchOptions(options, extended):
    return g.eval_search.convert_form_parameters(options,
                                                 session['USER_AUTHORIZED'],
                                                 extended=extended)

def search_options_from_session():
    options = session.get('AGENCY_OVERVIEW_SEARCHOPTIONS')
    if options:
        return options

    user = session['USER_AUTHORIZED']
    if user.isAdmin():
        return g.mpuls_config.get('search', 'admin_defaults')[0]
    else:
        return g.mpuls_config.get('search', 'cm_defaults')[0]


class EvaluationOverviewController(BaseController):

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def index(self):
        return self.overview()

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def overview_extended(self):
        extended = session.get('EXTENDED_SEARCH')
        if extended is not None:
            extended = not extended
        else:
            extended = True
        session['EXTENDED_SEARCH'] = extended
        session.save()
        return self.overview()

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def overview(self):
        # If reset parameter is set ti != 0 the delete all saved params and
        # cases on overview page
        reset = self._checkBool(request.params.get('reset', '0'))
        if reset:
            session['AGENCY_OVERVIEW_SEARCHOPTIONS'] = {}
            session['EXTENDED_SEARCH'] = False
            session.save()

        search_options = search_options_from_session()
        search_options.update(self.parseSortOptions(request.params))
        session['AGENCY_OVERVIEW_SEARCHOPTIONS'] = search_options
        session.save()

        return self.search(search_options)

    def _getCaseList(self, search_options):
        cases = MpulsAgencyOverview()
        if search_options:
            options = self.parseSearchOptions(search_options,
                                        extended=session.get("EXTENDED_SEARCH"))
            cases.search(options)
        return cases

    def _renderOverview(self, cases, form_results, form_defaults, form_errors):
        user = session['USER_AUTHORIZED']

        # Values used in the template
        c.extended_search = session.get('EXTENDED_SEARCH', False)
        ulo = UserListObject()
        filter = [u.id for u in ulo.getAdminList()]
        c.userlist = ulo.getUserList(filter)
        c.phases = PhaseFactory().build()

        #count num of cases
        c.count_agency = len([x for x in cases.getDatasets()])
        c.count_all = sum([int(x.count) for x in cases.getDatasets()])

        # Hide evaluation option if someone did a search for a particular editor
        c.hide_evaluation = False
        if form_results.get('editor') and form_results.get('editor') != -1:
            c.hide_evaluation = True

        session["id_field_candidate"] = "m.fkz::bigint"
        all_ids = " ".join(unicode(case.id)
            for case in cases.getDatasets() if case.id is not None)
        form_defaults["all_ids"] = all_ids
        session["sql_where_candidate"] = cases.getWhereSQL()
        session.save()

        c.cases = cases
        c.datadump = self._get_datadump()
        overview = render('/evaluation/overview.mako')
        return formencode.htmlfill.render(overview, defaults=form_defaults,
                                          errors=form_errors)


    @checkRole(('admin', 'cm', 'pt_dlr'))
    def search(self, search_options=None):
        validator = g.eval_search.get_form_validator()

        form_result = {}
        form_errors = {}
        form_defaults = None

        if search_options is None:
            search_options = {}
            form_result.update(search_options_from_session())
        else:
            form_result.update(search_options)

        if not search_options:
            try:
                form_result = validator.to_python(request.params)
                search_options.update(form_result)
                session['AGENCY_OVERVIEW_SEARCHOPTIONS'] = form_result
                session.save()
            except formencode.Invalid, error:
                log.exception("search: formencode error %s" % error)
                form_defaults = error.value
                form_errors = error.error_dict or {}

        if form_defaults is None:
            form_defaults = validator.from_python(form_result)

        cases = self._getCaseList(search_options)
        return self._renderOverview(cases, form_result, form_defaults,
                                    form_errors)

    def parseSortOptions(self, options=None):
        sort_field, sort_order = get_sort(options)

        # Pass the sort options to the template
        c.sort_field = sort_field
        c.sort_order = sort_order

        # default behavior is to take the sortfield for sorting
        return dict(sort_field=sort_field,
                    sort_order=sort_order)

    def parseSearchOptions(self, options, extended):
        return parseSearchOptions(options, extended)

    def _empty_caselist(self):
        c.url_ok = url(controller='evaluation_overview', action='index')
        c.dialog_title = _("No valid agencys in the selection!")
        c.dialog_text = _(u"""The selected action can not be executed
on any of the agencys in your selection. Please make sure that the agencys are
in a state from which you can execute the selected action.""")
        return render('/casebundle/dialogs/error.mako')

    def no_action(self):
        c.dialog_title= _("No action specified for the selection")
        c.dialog_text = _("""You have not selected any action for the
selection. Please click OK to go to the case overview and select there
an action from the drop down menu.""")
        c.url_ok = url(controller='evaluation_overview', action='index')
        return render('/casebundle/dialogs/error.mako')

    def bundleAction(self):
        validator = BundleActionForm()
        log.debug(request.params)
        try:
            # Check submitted fields
            form_result = validator.to_python(request.params)
            log.debug(form_result)

            # Check if the user has selected any cases
            if (len(form_result.get('case_id', [])) <= 0
                and not form_result.get('all_cases')):
                return self._empty_caselist()

            # Delegate to concrete action
            action = form_result.get('action')
            if action == 'no_action':
                return self.no_action()
            if action == 'evaluate':
                session["evaluation_ids"] = form_result.get('case_id')
                session["id_field"] = session.get("id_field_candidate")
                session["sql_where"] = session.get("sql_where_candidate")
                if form_result.get('all_cases') > 0:
                    session["evaluation_ids"] = form_result.get('all_ids')
                session.save()
                c.url_ok = url(controller='evaluation_overview', action='index')
                num = len(session["evaluation_ids"])
                c.dialog_title= ungettext("Agency was marked for evaluation",
                                          "Agency were marked for evaluation",
                                          num)
                c.dialog_text = \
                    ungettext("%d agency was marked for evaluation."
                              " All following evaluations will be limited"
                              " to that agency. Click 'OK' to continue.",
                              "%d agencys were marked for evaluation."
                              " All following evalutations will be limited"
                              " to those agencys. Click 'OK' to continue.",
                              num) % num

                return render('/casebundle/dialogs/success_markevaluate.mako')
        except formencode.Invalid, error:
            log.exception(error)
            raise
        except Exception, e:
            log.exception(e)
            raise
        return "Not implemented"

    def _get_datadump(self):
        v_dataset = None
        try:
            try:
                conn = db.getConnection()
                cur = conn.cursor()
                cur.execute("SELECT value FROM dbinfo WHERE key = 'datadump'")
                datestr = cur.fetchone()[0]
                y,m,d = datestr.split('-')
                v_dataset = date(int(y), int(m), int(d))
            except:
                log.exception('Error: get datadump failed')
        finally:
            db.recycleConnection(conn, None)
        return v_dataset

# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8:
