# -*- coding: utf-8 -*-
#
# (c) 2008 by Intevation GmbH
# This is Free software under the GPLv3.
# See LICENSE comming with the source of 'mpuls'
# for details.
#
# author: Torsten Irländer <torsten.irlaender@intevation.de>
#

import os.path
import datetime
import logging

import simplejson as json

from paste.reloader import watch_file

import pylons


log = logging.getLogger(__name__)


def get_revision():
    return "2.4.3"


def get_path(dir, file):
    try:
        # Return path of the first (usually user defined location) of public
        # files.
        return os.path.join(pylons.config.get('pylons.paths')
                            .get('static_files')[0], dir, file)
    except:
        return None # path building will fail on initial server start/restart
                    # as pylonspath is empty

def merge(a, b):
    """Recursively update the items of a with the corresponding values of b.

    Both a and b should be dictionaries.  All values of a which have
    corresponding values in b, i.e. values for the same keys, are merged
    into a.  If the values in b and a are dictionaries, they will be
    merged recursively, otherwise, the value of be will be inserted into
    a as the new value.

    The recursive update of dictionaries makes it difficult to have
    dictionaries as values where the user should be able to add keys.
    Simply using a dictionary object that is not a direct instance of
    dict works around the recursive merge.  The unmergeable_dict class
    can be used to create instances like this.
    """
    changes = []
    for ka, va in a.iteritems():
        try:
            vb = b[ka]
        except KeyError, err:
            continue
        if type(vb) == dict and type(va) == dict:
            merge(va, vb)
        elif vb != va:
            changes.append((ka, vb))
    a.update(changes)


class unmergeable_dict(dict):

    """Dict-equivalent for default values that should not be merged recursively.
    """


class MpulsConfig(object):

    def __init__(self, filename=None):
        self.config = self.build_defaults()
        if filename and os.path.isfile(filename):
            f = open(filename, "rb")
            try:
                log.info("Loading configfile: %s" % filename)
                cfg = json.load(f)
                log.info("Loading OK")
                # when running under the paste server, automatically
                # restart if the file changes.
                watch_file(filename)
            finally:
                f.close()
            merge(self.config, cfg)

    def build_defaults(self):
        """Create the initial settings dictionary with default values.
        The default implementation simply returns an empty dictionary.
        Derived classes should override this method to provide the
        appropriate default values.
        """
        return {}

    def get(self, section, name):
        try:
            return self.config[section][name]
        except KeyError:
            log.exception('Error on checking config (%s, %s)' % (section, name))
            return ''


class MpulsAppConfig(MpulsConfig):

    def build_defaults(self):
        """Build the default settings for an MPuls application"""

        sections = {}

        def set(section_key, item_key, item):
            try:
                section = sections[section_key]
            except KeyError:
                section = {}
                sections[section_key] = section
            section[item_key] = item

        set('common', 'name', 'wasko')
        set('common', 'version', 'rev. %s' % get_revision())
        set('common', 'releasedate', '25.09.2013')
        set('common', 'showdemonotice', '0')
        set('common', 'showagencychooser', '0')

        # Formed
        set('formed', 'treefile', 'formedtree.xml')
        set('formed', 'annotation', 'annotations.xhtml')

        # Offline-Client 
        set('offline-client', 'enabled', '0')

        # Global Modules
        set('module', 'administration', '1')
        set('module', 'casemanagement', '1')
        set('module', 'appointments', '1')
        set('module', 'documents', '1')
        set('module', 'documents-description', '0')
        set('module', 'news', '1')
        set('module', 'evaluation', '1')
        # Used for evaluation server.
        set('module', 'agency-overview', '0')
        # Used for the reportquota
        set('module', 'meta-transmit-reportquota', '0')

        # wether the reportquota view is enabled.
        set('module', 'meta-show-reportquota', '0')

        # Case Modules
        set('case-module', 'search', '1')
        set('case-module', 'casebundle', '1')
        set('case-module', 'appointments', '1')
        set('case-module', 'create', '1')
        set('case-module', 'import', '1')
        set('case-module', 'exportXML', '1')
        set('case-module', 'exportXLS', '1')
        set('case-module', 'exportCSV', '1')
        set('case-module', 'reminders', '1')
        set('case-module', 'privacy', '1')
        set('case-module', 'anonymize', '1')
        set('case-module', 'documents', '1')
        set('case-module', 'logbook', '1')
        set('case-module', 'organisation', '1') #set editors, standin
        set('case-module', 'tags', '0')
        set('case-module', 'statements', '0') # show links for statements under further options.
        #enable application based tagging (devel)
        set('case-module', 'checktags', '0') 
        #enable the case-copy function in template (devel)
        set('case-module', 'copy', '0') 
        #enable function to extract certain fields for formletters 
        set('case-module', 'formletter', '0') 

        # whether the meta-server support is enabled.
        set('case-module', 'meta-server', '0')


        # CASE FIELDS
        #
        # This defines which of the fields of a case need to be handled
        # specially in some way.  There's one dict for each such field.
        # The meaning of the settings for each field is described in the
        # doc-string of the Field class in mpulsweb/model/case.py.
        #
        # Some points are not described there, though, because they
        # apply to the list as a whole or specific fields with a special
        # meaning in mpulsweb:
        #
        #  - the order of the descriptions is relevant for the default
        #    case digest template which shows the fields in the order
        #    listed here.
        #
        #  - some fields are accessed via english alias names in the
        #    code.  These and the corresponding formed name in wasko
        #    are:
        #
        #       formed name        alias
        #      ----------------------------------
        #       name               last_name
        #       vorname            first_name
        #       fn                 knr
        #       erstgespraech      first_meeting
        #       datum_cm_ende      cm_end
        #       bearbeiter_id      editor
        #
        #  - Some fields are not actually fields in the formed
        #    description and need to be defined here or in the .json
        #    file with "preload":True and "search_retrieve":True :
        #
        #      bearbeiter_id
        #
        #  - alias_force_string should be set to False for all fields
        #    that have an alias and which are not strings, such as
        #    erstgespraech, datum_cm_ende which are dates, or
        #    bearbeiter_id which is an int.
        #
        set('case', 'fields', [{"name": "id",
                                "alias": "id",
                                "preload": True,
                                "overview": True,
                                "search_retrieve": True,
                                "search_match": True,
                                "overview_style": "table_width_small",
                                "overview_label": "ID"},
                               {"name": "bearbeiter_id",
                                "search_retrieve": True,
                                "preload": True,
                                "alias": "editor",
                                "alias_force_string": False}
                               ])
        set('agency', 'fields', [{"name":"fkz", 
                                  "alias":"fkz", 
                                  "preload":True, 
                                  "overview":True, 
                                  "search_retrieve":True, 
                                  "search_match":True, 
                                  "overview_style":"table_id", 
                                  "overview_label": "ID"},
                                 {"name":"id", 
                                  "alias":"id", 
                                  "preload":True},
                                 {"name":"name", 
                                  "alias":"name", 
                                  "preload":True, 
                                  "overview":True, 
                                  "search_retrieve":True, 
                                  "search_match":True, 
                                  "overview_style":"", 
                                  "overview_length":100, 
                                  "overview_label": "Name"},
                                 {"name":"count", 
                                  "alias":"count", 
                                  "preload":True, 
                                  "overview":True, 
                                  "overview_style":"table_width_mid_s number_field", 
                                  "overview_label": "Anzahl"},
                                 {"name":"time", 
                                  "alias":"time", 
                                  "preload":True, 
                                  "overview":True, 
                                  "overview_style":"table_date number_field", 
                                  "overview_label": "Aktualisiert"}])

        # On default MaxSaveTimeReminders will be generated after x-days from
        # last accesstime (defined in the agency configuration). Here you can
        # define another condition which must evaluate to TRUE or FALSE in SQL
        # manner. 
        set('case', 'remindercondition', 'TRUE')
        set('case', 'reminderconditiondesc', 'Keine')
        set('case', 'remindermindays', '0')
        set('case', 'remindermaxdays', '9999')

        # Set default values for an item  on autocomplete a case
        # Date can be set to "NOW" which is then replaced with the current date
        # or an dd.mm.jjjj formated datestring 
        set('case', 'default-autocomplete-date', 'unbekannt')
        set('case', 'default-autocomplete-integer', '-9999999')
        set('case', 'default-autocomplete-text', 'unbekannt')
        set('case', 'default-autocomplete-choice', '-3')

        # JOB PERCENTAGE
        # --------------
        # Configure the values for the job percentage
        set('job_percentage', 'enabled', 0)
        set('job_percentage', 'start_date', '2011-09-01')
        set('job_percentage', 'months', '16')
        set('job_percentage', 'threshold_lower', 0)
        set('job_percentage', 'threshold_higher', 0)
        set('job_percentage', 'from_email', '0')
        set('job_percentage', 'to_email', '0')
        set('job_percentage', 'phases' , '1,2,3')
        set('job_percentage', 'recipient', '0')
        set('job_percentage', 'showeval', '0')


        # FORMLETTER 
        # ----------
        # Defines a list of different formletters. Each letter has a id and a
        # name which is used in the GUI. Earch sql-querys of a formletter will
        # collect the needed data for the formletter. All querys are done one
        # after another and its result will be written in a dictionary.
        # SQL querys can include %(id)s which will be substituted with the id
        # of the case. 
        sqllist = []
        letters = [{'id': '1', 
                    'name': 'Vorgabe', 
                    'sql': [
                        'select * from master_tbl_view where id = %(id)s'
                    ]}]
        set('case', 'formletter', letters)

        # IMPORTANT TAGS
        # --------------
        #
        # Specify here a list of "important" tags. For cases with these tags,
        # reminders will be shown in the overview, so use sparingly. The
        # description of the reminder will be the one from the tag, as
        # specified in the XML file.
        set('overview_tags', 'tag_ids', '')

        # PHASES
        # --------
        # Mode of phase changes
        # Set the mode how phases are changed
        #   * 0 = Manual, old behavior (Default)
        #   * 1 = Half automatic, Dialog based
        #   * 2 = Full automatic
        set('phases', 'changemode', '0')

        # Hide "unknown" phase in GUI. The unknown phase is part of the
        # phasemodel but we very likley do not want it to be displayed in the
        # GUI. So set this value to 1 on default  
        set('phases', 'hideunknown', '1')

        # Depending on the set phasepart the case is considered as CM-Case or Beratungs-Case.
        # Here you can define the type of case for each phasepart
        set('phases', 'casetype', [{}])

        # Available Phases:
        set('phases', 'phases', ['-1'])

        # Description of phases
        description = [{'-1': 'Unbekannt'}]
        set('phases', 'description', description)
        shortdescription = [{'-1': 'U'}]
        set('phases', 'shortdescription', description)
        icons = [{'-1': 'u'}] # letter will be exanded with "running" or "stopped"
        set('phases', 'icons', icons)

        # Phasepairs
        pairs =[{"-1": ["-1", "-1"]}]
        set('phases', 'pairs', pairs)

        # Dates 
        dates =[]
        set('phases', 'dates', dates)

        # Not anonymizable phaseparts
        set('phases', 'notanonymizable', [])

        # Set condition which needs to to be true to check only the minimal
        # fields for anonymisation instead of all required fields. Notation is
        # the same used in rules in formed (polnish notation). Default is "0" 
        # which means the on default the case must be in full consistence state
        # before it is allowed to be anonymized
        set('phases', 'minimal_consistence_conditon', "0")

        # Set minimal fields which must be set before anonymize
        set('phases', 'minimal_fields_to_anonymize', [])

        # Succenssors and Predesessors
        successors = [{"-1": []}]
        set('phases', 'successors', successors)

        # The following lines configure in which phases a case is considered as
        # active or finished. This is used in determining outaged cases (e.g
        # max savetime)
        # Which phases are considered as active
        active = ["-1"]
        set('phases', 'active', active)

        # Which phases are considered as finished
        finished = ["-1"]
        set('phases', 'finished', finished)

        # SEARCH
        # ------
        # Admin defaults:
        search_admin_default = [
            {
                'own': 0,
                'standin': 0,
                'editor': -1,
                'search_str': '',
                'state': [3, 4],
                'phase': [-1],
                'federal_state': [0, 1],
                }
            ]
        set('search', 'admin_defaults', search_admin_default)

        # CM defaults:
        search_cm_default = [
            {
                'own': 1,
                'standin': 0,
                'editor': -1,
                'search_str': '',
                'state': [1],
                'phase': [-1],
                'federal_state': [0, 1],
                }
            ]
        set('search', 'cm_defaults', search_cm_default)

        set('search', 'default-sorting', 'id')
        set('search', 'default-order', 'asc')
        set('search', 'default-start-date', "2011-09-01")
        set('search', 'default-end-date', "2012-08-31")

        # Enable an disable certain search-options
        set('search', 'option-state', '1')
        set('search', 'option-editor', '1')
        set('search', 'option-responsibilty', '1')
        set('search', 'option-agency', '1')
        set('search', 'option-phase', '1')
        set('search', 'option-date', '1')
        set('search', 'option-gender', '1')
        set('search', 'option-federalstate', '0')

        # LOGBOOK
        # -------
        # Available types of entries
        set('logbook', 'descriptions', [{'-1': 'Keine Angabe', 
                                         '-2': 'Sonstige Angaben'}])
        set('logbook', 'categories', [{'name': 'Alle', 
                                       'items': ['-1', '-2']}])
        set('logbook', 'types', [{'db_field': 'art', 
                                  'label': 'Art des Eintrags', 
                                  'options': [{'label': 'Alle', 
                                               'options': [['-1', 'Keine Angabe'], 
                                                           ['-2', 'Sonstige Angaben']]}]}])

        #  EVALUATIONS
        # ------------
        #
        # Note the the evaluation with id 0 is considered to be the
        # Adele-evaluation
        dummy = {'name': 'Vorgabe', 'id': '99', 'config': 'eval_default.xml', 'options_disabled': '0'}
        enabled_evaluations = [dummy]
        set('evaluations', 'enabled', enabled_evaluations)
        set('evaluations', 'adele-phases', [])
        set('evaluations', 'default-phases', [])
        set('evaluations', 'default-ending', [])

        # VALIDITIES
        # -------------
        # Note that the grade of validity is defined by the ascending order of
        # the elements in enabled_validities list. First element has lowest
        # validity, last element has higest validity 
        validity_0 = {'name': 'Nicht auswertbar', 
                      'id': '0', 
                      'tags': [], 
                      'icon': 'validity-0-22.png'}
        validity_1 = {'name': 'Grundsätzlich auswertbar', 
                      'id': '1', 
                      'tags': [], 
                      'icon': 'validity-1-22.png'}
        validity_2 = {'name': 'Berichtsfähig', 
                      'id': '2', 
                      'tags': [], 
                      'icon': 'validity-2-22.png'}
        enabled_validities = [validity_0]
        set('validities', 'enabled', enabled_validities)

        return sections

    def is_enabled(self, section, name):
        try:
            return self.get(section, name) == '1'
        except:
            log.exception('Error on checking config (%s, %s)' % (section, name))
            return False

    def get_date(self, section, name):
        """Return the value for section and name interpreted as a date.
        The value should either be an empty string in which case None is
        returned or a string in ISO format (YYYY-MM-DD) which is
        converted to a date object.  If the given configuration setting
        doesn't exist, None is returned.
        """
        value = self.get(section, name)
        if value:
            return datetime.date(*[int(part) for part in value.split("-")])
        else:
            return None

    def get_app_name(self):
        return self.get('common', 'name')

    def get_app_version(self):
        return self.get('common', 'version')

    def get_releasedate(self):
        return self.get('common', 'releasedate')

    def get_overview_fields(self):
        return list(self.get('case-fields', 'overview'))

    def get_fieldname(self, key, mapping='case-fields-mapping'):
        '''Returns the mapped name of database field for a given mpuls internal
        datafield. If the field is not mapable return the key'''
        if not mapping:
            mapping = 'case-fields-mapping'
        try:
            return self.config[mapping][key]
        except KeyError:
            return key


class MpulsDBMapping(MpulsConfig):

    def build_defaults(self):
        """Build the default settings for the DB mapping"""

        sections = {}

        def set(section_key, item_key, item):
            try:
                section = sections[section_key]
            except KeyError:
                section = {}
                sections[section_key] = section
            section[item_key] = item
        mapping = {}
        set('db', 'mapping', [mapping])
        return sections

    def get_mapping(self):
        mapping = self.get('db', 'mapping')
        if mapping != '':
            return mapping[0]
        else: return {}

class MpulsDBList(MpulsConfig):

    def build_defaults(self):
        """Build a list of available DB"""

        sections = {}

        def set(section_key, item_key, item):
            try:
                section = sections[section_key]
            except KeyError:
                section = {}
                sections[section_key] = section
            section[item_key] = item
        dblist= {}
        set('db', 'list', [dblist])
        return sections

    def get_list(self):
        dblist= self.get('db', 'list')
        if dblist != '':
            return dblist[0]
        else: return {}

    def get_field(self, name, fieldname):
        try:
            res1 = self.get_list()[name]
            field = res1[fieldname]
        except KeyError:
            return name
        return field

# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8:
