# -*- coding: utf-8 -*-
import logging
import datetime
import traceback

from mpulsweb.lib.db import db
from pylons import app_globals as g

log = logging.getLogger(__name__)


### case count function for report quota
#SQL
SQL_COUNT_CASES = """\
SELECT COUNT(*) FROM master_tbl_reportquota_view WHERE TRUE"""

SQL_COUNT_CASES_SYNC = """\
SELECT COUNT(*) FROM master_tbl_reportquota_view WHERE sync = true"""

#Errors
class CountCaseError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)

def load_count_cases():
    count_cases = None
    count_cases_sync = None
    conn, cur = None, None
    try:
        conn = db.getConnection()
        cur = conn.cursor()

        query = SQL_COUNT_CASES
        log.debug("Query: %s", query)
        cur.execute(query)
        count_cases = cur.fetchone()[0]

        query = SQL_COUNT_CASES_SYNC
        log.debug("Query: %s", query)
        cur.execute(query)
        count_cases_sync = cur.fetchone()[0]
    except:
        log.exception("load_count_cases: failed")
    finally:
        db.recycleConnection(conn, cur)

    if not count_cases is None and not count_cases_sync is None:
        if count_cases < count_cases_sync:
            raise CountCaseError("count_case_sync larger than count_case")
        return (count_cases, count_cases_sync)
    else:
        raise CountCaseError("Count cases failed")

STORE_REPORTQUOTA_SQL ="""\
INSERT INTO berichtsquote_tbl
       (anzahl_akten, anzahl_akten_sync, fkz)
VALUES (%(count_cases)s, %(count_cases_sync)s, %(fkz)s)"""

class ReportQuotaImporter:

    def store_rq(self, dict1):
        if "count_cases" in dict1 and "count_cases_sync" in dict1 \
        and "fkz" in dict1:
            conn = None
            try:
                try:
                    conn = db.getConnection()
                    cur = None
                    try:
                        cur = conn.cursor()
                        cur.execute(STORE_REPORTQUOTA_SQL, dict1)
                        conn.commit()
                    finally:
                        if cur:
                            try:
                                cur.close()
                            except:
                                pass
                finally:
                    db.recycleConnection(conn)
            except:
                log.exception("Store report quota failed")
        else:
            log.error("Not complete arguments, cannot create rq entry.")

# need startyear and endyear
LOAD_BERICHTSQUOTE_SQL = """\
SELECT upload, anzahl_akten, anzahl_akten_sync, fkz FROM berichtsquote_tbl
WHERE date_part('year',upload) >= %(startyear)s
AND date_part('year', upload) <= %(endyear)s
ORDER BY upload ASC"""

LOAD_FKZS_SQL = """\
SELECT fkz FROM berichtsquote_tbl"""

class ReportQuotaEntry(object):
    """Handle the data to calculate a rq and the update."""

    def __init__(self, update, count_cases, count_cases_sync):
        self.update = update
        self.count_cases = count_cases
        self.count_cases_sync = count_cases_sync
        if self.count_cases == 0:
            self.quota = 0.0
        else:
            self.quota = float(self.count_cases_sync) / float(self.count_cases)

class ReportQuotaData(object):

    def __init__(self, year):
        self.reportquota_dict = {}
        self.fkzs = []
        self._load_all_fkzs()
        self._load(year, year)
        self.year = year

    def _load(self, startyear, endyear):
        result = None
        conn, c = None, None
        try:
            conn = db.getConnection()
            c = conn.cursor()
            query = LOAD_BERICHTSQUOTE_SQL
            fields = {'startyear': startyear, 'endyear': endyear}
            log.debug("Query: %s Fields: %s", query, str(fields))
            c.execute(query, fields)
            result = c.fetchall()
        finally:
            db.recycleConnection(conn, c)

        months = range(1,13)
        years   = range(startyear, endyear+1)

        if result:
            for y in years:
                for m in months:
                    for r in result:
                        self._modify_dict(y, m, r)

    def _load_all_fkzs(self):
        result = None
        conn, c = None, None
        try:
            conn = db.getConnection()
            c = conn.cursor()
            query = LOAD_FKZS_SQL
            log.debug("Query: %s", query)
            c.execute(query)
            result = c.fetchall()
        finally:
            db.recycleConnection(conn, c)
        for r in result:
            if not r[0] in self.fkzs:
                self.fkzs.append(r[0])

    def _modify_dict(self, y, m, r):
        r_update = r[0]
        r_fkz = r[3]
        new_reportquotaentry = ReportQuotaEntry(r_update, r[1], r[2])
        if r_update.year <= y and r_update.month <= m:
            today = datetime.date.today()
            # don't extrapolate into the future.
            if today.year > y or today.year == y and today.month >= m:
                self.reportquota_dict[(y,m,r_fkz)] = new_reportquotaentry

    def _print_res_dict(self):
        d = self.reportquota_dict
        buf = []
        for (y,m,f) in d:
            buf.append("%s %s %s: %s %s" % (y, m, f,
                            d[(y,m,f)].update,
                            d[(y,m,f)].quota))
        return "\n".join(buf)

    def get(self, x, default1):
        return self.reportquota_dict.get(x, default1)

    def get_fkzs(self):
        return sorted(self.fkzs)

    def get_x_count_cases(self, fkz, report_range):
        count = 0
        sum1 = 0
        for (y,m) in report_range:
            rqe = self.get((y,m,fkz),None)
            if rqe:
                count += 1
                sum1 +=  rqe.count_cases
        if count == 0:
            return 0.0
        else:
            return float(sum1)/float(count)

    def get_x_count_cases_sync(self, fkz, report_range):
        count = 0
        sum1 = 0
        for (y,m) in report_range:
            rqe = self.get((y,m,fkz),None)
            if rqe:
                count += 1
                sum1 +=  rqe.count_cases_sync
        if count == 0:
            return 0.0
        else:
            return float(sum1)/float(count)

    def get_x_quota(self, fkz, report_range):
        cc = self.get_x_count_cases(fkz, report_range)
        ccs = self.get_x_count_cases_sync(fkz, report_range)
        if abs(cc) < 0.00001:
            return 0.0
        else:
            return ccs/cc

    def get_x_dates(self, fkz, report_range):
        date_max = None
        for (y,m) in report_range:
            rqe = self.get((y,m,fkz),None)
            if rqe:
                if not date_max or date_max < rqe.update:
                    date_max = rqe.update
        if date_max:
            return "%s.%s." % (date_max.day, date_max.month)
        else:
            return "-/-"

    def get_ysum_count_cases(self, y, m):
        sum1 = 0
        for fkz in self.fkzs:
            rqe = self.get((y,m,fkz),None)
            if rqe:
                sum1 += rqe.count_cases
        return sum1

    def get_ysum_count_cases_sync(self, y, m):
        sum1 = 0
        for fkz in self.fkzs:
            rqe = self.get((y,m,fkz),None)
            if rqe:
                sum1 += rqe.count_cases_sync
        return sum1

    def get_ysum_quota(self,y, m):
        cc = self.get_ysum_count_cases(y,m)
        ccs = self.get_ysum_count_cases_sync(y,m)
        if cc == 0:
            return 0.0
        else:
            return  float(ccs)/float(cc)

    def get_ysum_date(self, y, m):
        date_max = None
        for fkz in self.fkzs:
            rqe = self.get((y,m,fkz),None)
            if rqe:
                if not date_max or date_max < rqe.update:
                    date_max = rqe.update
        if date_max:
            return "%s.%s." % (date_max.day, date_max.month)
        else:
            return "-/-"

    def get_program_name_from_fkz(self, fkz):
        return g.mpuls_dblist.get_field(fkz, "measure_name")

    def get_rb_case_count(self, year_list):
        case_count = 0
        for (y,m) in year_list:
            case_count += self.get_ysum_count_cases(y,m)
        return case_count

    def get_rb_case_count_sync(self, year_list):
        case_count_sync = 0
        for (y,m) in year_list:
            case_count_sync += self.get_ysum_count_cases_sync(y,m)
        return case_count_sync

    def get_rb_quota(self, year_list):
        cc = self.get_rb_case_count(year_list)
        ccs = self.get_rb_case_count_sync(year_list)
        if cc == 0:
            return 0.0
        else:
            return  float(ccs)/float(cc)

    def get_rb_date(self, year_list):
        date_max = None
        for (y,m) in year_list:
            for fkz in self.fkzs:
                rqe = self.get((y,m,fkz),None)
                if rqe:
                    if not date_max or date_max < rqe.update:
                        date_max = rqe.update
        if date_max:
            return "%s.%s." % (date_max.day, date_max.month)
        else:
            return "-/-"
