# -*- coding: utf-8 -*-
#
# Copyright 2007, 2008 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.
#
# Authors:
# Sascha L. Teichmann <teichmann@intevation.de>
# Bernhard Herzog <bernhard.herzog@intevation.de>
#

from __future__ import with_statement

import base64
from mimetypes import guess_type
import xml.etree.ElementTree as ET

import psycopg2  as dbapi

from formed.meta.data import RepeatNode, TextLeaf, DateLeaf
from formed.meta.structure import StructureNode
from formed.instance.backends.xmlimport import ComponentFactory
from formed.instance.backends.xmlexport import add_leaf
from formed.instance.simplified import StructureInstanceNode, \
     StructureInstanceTree

import mpulsweb.lib.helpers as h
from mpulsweb.lib.base import g
from mpulsweb.lib.db import db
from mpulsweb.lib.translation import _
from mpulsweb.model.logbook import text_subelement, new_node

MAX_SIZE = 10 * 1024 * 1024
BLOCK_SIZE = 4 * 1024

CREATE_GLOBAL = 'SELECT * FROM create_global_dokument_ds(%(name)s)'
CREATE_CASE = '\
SELECT * FROM create_fall_dokument_ds(%(id)s, %(uuid)s, %(name)s)'

DELETE_GLOBAL = '''
SELECT delete_global_dokument_ds(%(id)s)
'''
DELETE_CASE = '''
SELECT delete_fall_dokument_ds(%(id)s)
'''

    
UPDATE_GLOBAL_WITH_DESC = '''
UPDATE ka_global_dokumente_tbl_view
SET daten = %(bytes)s, groesse = %(size)s, name = %(name)s,
mime = %(mime)s, beschreibung = %(description)s, datum = %(datum)s
WHERE id = %(id)s
'''

LIST_GLOBAL_WITH_DESC = '''
SELECT id, master_id, groesse, mime, name, datum, beschreibung
FROM ka_global_dokumente_tbl_view
'''

LOAD_GLOBAL_WITH_DESC = '''
SELECT master_id, groesse, mime, daten, name, datum, beschreibung
FROM ka_global_dokumente_tbl_view
WHERE id = %(id)s
'''

RENAME_GLOBAL_WITH_DESC = '''
UPDATE ka_global_dokumente_tbl_view
SET name = %(name)s, beschreibung = %(desc)s
WHERE id = %(id)s
'''

UPDATE_GLOBAL = '''
UPDATE ka_global_dokumente_tbl_view
SET daten = %(bytes)s, groesse = %(size)s, name = %(name)s, mime = %(mime)s,
datum = %(datum)s
WHERE id = %(id)s
'''

RENAME_GLOBAL = '''
UPDATE ka_global_dokumente_tbl_view
SET name = %(name)s
WHERE id = %(id)s
'''

LIST_GLOBAL = '''
SELECT id, master_id, groesse, mime, name, datum
FROM ka_global_dokumente_tbl_view
'''

LOAD_GLOBAL = '''
SELECT master_id, groesse, mime, daten, name, datum
FROM ka_global_dokumente_tbl_view
WHERE id = %(id)s
'''

UPDATE_CASE_WITH_DESC = '''
UPDATE ka_fall_dokumente_tbl_view
SET daten = %(bytes)s, groesse = %(size)s, name = %(name)s,
mime = %(mime)s, beschreibung = %(description)s, datum = %(datum)s
WHERE id = %(id)s AND master_id = %(case)s
'''

LIST_CASE_WITH_DESC = '''
SELECT id, master_id, groesse, mime, name, datum, beschreibung
FROM ka_fall_dokumente_tbl_view
WHERE master_id = %(master_id)s ORDER BY datum DESC, name
'''

LOAD_CASE_WITH_DESC = '''
SELECT master_id, groesse, mime, daten, name, datum, beschreibung
FROM ka_fall_dokumente_tbl_view
WHERE id = %(id)s
'''

RENAME_CASE_WITH_DESC = '''
UPDATE ka_fall_dokumente_tbl_view
SET name = %(name)s, beschreibung = %(desc)s
WHERE id = %(id)s
'''

UPDATE_CASE = '''
UPDATE ka_fall_dokumente_tbl_view
SET daten = %(bytes)s, groesse = %(size)s, name = %(name)s, mime = %(mime)s,
datum = %(datum)s
WHERE id = %(id)s AND master_id = %(case)s
'''

LIST_CASE = '''
SELECT id, master_id, groesse, mime, name, datum
FROM ka_fall_dokumente_tbl_view
WHERE master_id = %(master_id)s ORDER BY datum DESC, name
'''

LOAD_CASE = '''
SELECT master_id, groesse, mime, daten, name, datum
FROM ka_fall_dokumente_tbl_view
WHERE id = %(id)s
'''

RENAME_CASE = '''
UPDATE ka_fall_dokumente_tbl_view
SET name = %(name)s 
WHERE id = %(id)s
'''

class DocumentException(Exception):

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


def list_documents_with_cursor(cursor, case_id):
    documents = []

    if case_id is None:
        if g.mpuls_config.is_enabled("module", "documents-description"):
            stmnt = LIST_GLOBAL_WITH_DESC
        else:
            stmnt = LIST_GLOBAL
        cursor.execute(stmnt)
    else:
        if g.mpuls_config.is_enabled("module", "documents-description"):
            stmnt = LIST_CASE_WITH_DESC
        else:
            stmnt = LIST_CASE
        cursor.execute(stmnt, {'master_id': case_id})

    while True:
        row = cursor.fetchone()
        if not row:
            break
        if g.mpuls_config.is_enabled('module', 'documents-description'):
            id, case, size, mime, name, insert_date, description = row
        else:
            id, case, size, mime, name, insert_date = row
            description = None
        documents.append(Document(id, case, size, mime, name, description, 
                                  insert_date))

    return documents


def create_document_with_cursor(cursor, name, content, mime, case_id, uuid,
                                description="", insert_date='TODAY'):
    if case_id is None:
        cursor.execute(CREATE_GLOBAL, {'name': name})
    else:
        cursor.execute(CREATE_CASE,
                       {'id': case_id, 'uuid': uuid, 'name': name})

    row = cursor.fetchone()
    if row is None:
        raise DocumentException(_(u"Could not create new document"))

    document_id = row[0]
    fields = {'bytes': dbapi.Binary(content), 'size': len(content),
              'name': name, 'mime': mime, 'id': document_id,
              'description': description, 'datum': insert_date}

    if case_id is None:
        if g.mpuls_config.is_enabled("module", "documents-description"):
            stmnt = UPDATE_GLOBAL_WITH_DESC
        else:
            stmnt = UPDATE_GLOBAL
    else:
        if g.mpuls_config.is_enabled("module", "documents-description"):
            stmnt = UPDATE_CASE_WITH_DESC
        else:
            stmnt = UPDATE_CASE
        fields['case'] = case_id

    cursor.execute(stmnt, fields)

    return document_id

def rename_document_with_cursor(cursor, id, name, desc, is_case_document):
    if g.mpuls_config.is_enabled("module", "documents-description"):
        fields = {'id': id, 'name': name, 'desc':desc}
        if is_case_document:
            statement = RENAME_CASE_WITH_DESC 
        else:
            statement = RENAME_GLOBAL_WITH_DESC
    else:
        fields  = {'id': id, 'name': name}
        if is_case_document:
            statement = RENAME_CASE
        else:
            statement = RENAME_GLOBAL
    cursor.execute(statement, fields)


def delete_document_with_cursor(cursor, id, is_case_document):
    if is_case_document:
        statement = DELETE_CASE
    else:
        statement = DELETE_GLOBAL
    cursor.execute(statement, {'id': id})


def listDocuments(case=None):
    if not case is None:
        case = int(case)

    con, cur = None, None
    try:
        con = db.getConnection()
        cur = con.cursor()
        return list_documents_with_cursor(cur, case)
    finally:
        db.recycleConnection(con, cur)

def renameDocument(id, name, desc, case=False):
     with db.transaction() as cur:
        rename_document_with_cursor(cur, int(id), 
                                    unicode(name), unicode(desc),
                                    case)

def deleteDocument(id, case=False):
    with db.transaction() as cur:
        delete_document_with_cursor(cur, int(id), case)


class Document:

    def __init__(self, id=None, case=None, size=None, mime=None, name=None,
                       description=None, insert_date=None):
        self.id = id
        self.case = case
        self.size = size
        self.mime = mime
        self.name = name
        self.data = None
        self.description = description
        self.insert_date = insert_date

    def isCaseAttached(self):
        return not self.case is None

    def getName(self):
        name = self.name
        if name is None:
            return None
        return h.ensure_unicode(name)

    def create(self, name, src, case=None, uuid=None, description=None):
        name = name.strip().rsplit('\\', 1)[-1].rsplit('/', 1)[-1]
        if not name:
            raise DocumentException(_(u"No document available"))

        self.name = name

        mime = guess_type(name)
        if mime:
            mime = mime[0]
        else:
            mime = "application/octet-stream"

        self.mime = mime

        if not case is None:
            case = int(case)
            self.case = case

        bytes = src 
        self.size = len(bytes) #total_size

        with db.transaction() as cur:
            self.id = create_document_with_cursor(cur, name, bytes, mime, case,
                                                  uuid, description)

    def load(self, asCase=False):
        if self.id is None:
            raise DocumentException(_(u"An ID is needed"))
        con, cur = None, None
        try:
            con = db.getConnection()
            cur = con.cursor()
            if asCase:
                if g.mpuls_config.is_enabled("module", "documents-description"):
                    stmnt = LOAD_CASE_WITH_DESC
                else:
                    stmnt = LOAD_CASE
                cur.execute(stmnt, {'id': self.id})
            else:
                if g.mpuls_config.is_enabled("module", "documents-description"):
                    stmnt = LOAD_GLOBAL_WITH_DESC
                else:
                    stmnt = LOAD_GLOBAL
                cur.execute(stmnt, {'id': self.id})
            row = cur.fetchone()
            if not row:
                raise DocumentException(_(u"The requested document can't be found."))
            self.case = row[0]
            self.size = row[1]
            self.mime = row[2]
            self.data = row[3]
            self.name = row[4]
            self.insert_date = row[5]
            if g.mpuls_config.is_enabled("module", "documents-description"):
                self.description = row[6]
            else:
                self.description = None
        finally:
            db.recycleConnection(con, cur)


def export_documents(case, case_id, cursor, selection, anonymize):
    """Add the documents of the case with id case_id to the elementtree case"""
    documents_element = ET.SubElement(case, "anlagen")
    for document in listDocuments(case_id):
        document.load(True)
        document_element = ET.SubElement(documents_element, "anlage")
        add_leaf(document_element, "anlage-mime", document.mime)
        add_leaf(document_element, "anlage-name", document.name)
        add_leaf(document_element, "anlage-daten",
                 base64.b64encode(document.data))
        add_leaf(document_element, "anlage-beschreibung", document.description)
        add_leaf(document_element, "anlage-datum", document.insert_date)

def get_documents_import_factory():
    """Return component factory for documents for libformed xmlimport"""
    entry_node = StructureNode(new_node(RepeatNode, "anlage"),
                               [new_node(TextLeaf, "anlage-mime"),
                                new_node(TextLeaf, "anlage-name"),
                                new_node(TextLeaf, "anlage-daten"),
                                new_node(TextLeaf, "anlage-beschreibung"),
                                new_node(DateLeaf, "anlage-datum")
                                ],
                               [])
    return ComponentFactory(None,
                            StructureNode(None, [], [entry_node]),
                            StructureInstanceNode,
                            StructureInstanceTree)


def import_documents(case_tree, cursor, logbook_tree):
    """Import the documents"""
    documents = logbook_tree.root.get_repeat_groups("anlage")
    if documents:
        case_id = case_tree.getRootNode().getIdentifier()
        known = load_known_documents(cursor, case_id)
        for document in documents.itervalues():
            name = document.get_leaf("anlage-name")
            if name in known:
                del known[name]
            import_document(cursor, case_id, document)
        # delete documents which aren't imported again
        for doc in known.itervalues():
            delete_document_with_cursor(cursor, doc.id, True)


def import_document(cursor, case_id, document):
    name = document.get_leaf("anlage-name")
    mime = document.get_leaf("anlage-mime")
    data = document.get_leaf("anlage-daten")
    if name is None or mime is None or data is None:
        raise DocumentException(_("The document seems to be corrupted. Invalid "
                                  "anlage in document. Missing fields "
                                  "anlage-name, anlage-mime or anlage-daten."))
    content = base64.b64decode(data)
    description = document.get_leaf("anlage-beschreibung")
    insert_date = document.get_leaf("anlage-datum")
    create_document_with_cursor(cursor, name, content, mime, case_id, None,
                                description, insert_date=insert_date)


def load_known_documents(cursor, case_id):
    documents = list_documents_with_cursor(cursor, case_id)
    return dict((doc.name, doc) for doc in documents)


# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8:
