#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Authors:
#   Ludwig Reiter
#   Bernhard Herzog

"""Administrate mpuls databases.
Provides commands to init, create, upgrade and delete mpuls databases.

.. currentmodule:: mpulsdbadm
"""

import sys
import os
import os.path
import logging
import optparse
import re
import subprocess
import tempfile
import time

logging.basicConfig(level=logging.DEBUG)

REAL = 1


command_map = dict()

def register_command(name, command_class):
    command_map[name] = command_class


class CommandMetaClass(type):

    def __new__(mcs, name, bases, dict):
        cls = type.__new__(mcs, name, bases, dict)
        if cls.name is not None:
            register_command(cls.name, cls)
        return cls

def find_command(name):
    return command_map.get(name)

def list_commands():
    return sorted(command_map.items())


class Command(object):

    __metaclass__ = CommandMetaClass

    name = None

    def __init__(self):
        self._logging_steam = None

    def get_option_parser(self):
        parser = optparse.OptionParser(prog="%s %s" % (sys.argv[0], self.name))
        self.add_options(parser)
        return parser

    def add_options(self, parser):
        parser.add_option("-p", "--port", action="store", dest="port")
        parser.add_option("--stdout", action="store_true", dest="std_out",
                          default=False)
        parser.add_option("--dry-run", action="store_true", dest="dryrun",
                          default=False)
        parser.add_option("--log-dir", default="logs")
        parser.add_option("--script-dir")
        parser.add_option("--script-list")
        parser.add_option("--script-file", action="append",
                          dest="script_files")

    def check_options(self, options):
        global REAL
        if not options.port:
            print >>sys.stderr, "Need to specify a port number"
            return False
        if options.dryrun:
            REAL = 0
        return True

    def list_scripts(self, options):
        scripts = []
        if options.script_files:
            scripts.extend(options.script_files)
        if options.script_list:
            scripts.extend(file_to_list(options.script_list))
        if options.script_dir:
            scripts.extend(list_scripts(options.script_dir))
        return scripts

    def create_logdir(self, base_dir):
        directory = os.path.join(base_dir,
                                 (time.strftime("%Y%m%d-%H%M%S",
                                                time.localtime())
                                  + "-" + self.name))
        logging.info("Log to %r", directory)
        if not os.path.isdir(directory):
            os.makedirs(directory)
        return directory

    def logging_stream(self):
        return self._logging_stream

    def logging_handler(self, dbname, logging_dir):
        f = open(os.path.join(logging_dir, dbname + ".log"), "w")
        self._logging_stream = f
        handler = logging.StreamHandler(f)
        formatter = logging.Formatter(fmt=('%(asctime)s %(levelname)-8s'
                                           ' %(message)s'))
        handler.setFormatter(formatter)
        return handler

    def add_handler(self, handler):
        logging.getLogger().addHandler(handler)

    def remove_handler(self, handler):
        logging.getLogger().removeHandler(handler)

    def add_logging_handler(self, options, logging_dir, dbname):
        if options.std_out:
            logging.info("Log to stdout")
            return None

        handler = self.logging_handler(dbname, logging_dir)
        self.add_handler(handler)
        return handler

    def remove_logging_handler(self, handler):
        if handler is not None:
            self.remove_handler(handler)

    def run(self, options):
        self.execute(options)


class DBCommand(Command):

    def add_options(self, parser):
        super(DBCommand, self).add_options(parser)
        parser.add_option("-d", "--db", action="callback", type="string",
                          callback=single_argument_as_list, dest="dbnames",
                          metavar="DBNAME")
        parser.add_option("--dblist", action="callback", type="string",
                          callback=file_contents_as_list, dest="dbnames",
                          metavar="DBNAMEFILE")

    def check_options(self, options):
        valid = super(DBCommand, self).check_options(options)
        if not options.dbnames:
            print >>sys.stderr, "Need to specify a database name"
            valid = False
        return valid

    def full_db_name(self, agency):
        return "ka_" + agency + "_db"

    def execute(self, options):
        logging_dir = self.create_logdir(options.log_dir)
        for agency in options.dbnames:
            handler = self.add_logging_handler(options, logging_dir,
                                               self.full_db_name(agency))
            try:
                self.execute_for_db(options, agency)
            finally:
                self.remove_logging_handler(handler)

    def execute_for_db(self, options, agency):
        pass


class InitClusterCommand(Command):
    """
    Initialises a database cluster
    
    To create a new cluster use the following command::

	pg_createcluster 8.3 auswertung-wasku --locale de_DE.UTF-8

    If the cluster was new created it needs to be initialised to be ready to
    setup mpuls-database.
    This can be done using the follwing command::

	python mpulsdbadm.py init_cluster \\
		-p 5436 \\
		--script-file=/home/mpuls/wasku/db_config/db/mpuls-init-cluster.sql
		--log-dir=/tmp \\
    """

    name = "init_cluster"

    def check_options(self, options):
        valid = super(InitClusterCommand, self).check_options(options)
        if options.script_dir is None and os.path.isdir(self.name):
            options.script_dir = self.name
        return valid

    def execute(self, options):
        logging_dir = self.create_logdir(options.log_dir)
        handler = self.add_logging_handler(options, logging_dir, "cluster")
        try:
            self.check_connection(options)
            self.check_mpulscluster(options)
            self.init_cluster(options)
        finally:
            self.remove_logging_handler(handler)

    def check_connection(self, options):
        cmd = ["psql", "-p", options.port, "-c", r"\q"]
        logging.debug("command: %r", cmd)
        if REAL:
            exec_command(cmd)

    def check_mpulscluster(self, options):
        cmd = ["psql", "-p", options.port, "-c", r"\q", "mpuls_maintenance"]
        logging.debug("command: %r", cmd)
        if REAL:
            try:
                exec_command(cmd)
                failed = True
            except:
                failed = False
            if failed:
                logging.error("check_mpulscluster: Failed")
                raise Exception("Error DB mpuls_maintenance already exists"
                                " on port %r" % options.port)

    def init_cluster(self, options):
        scripts = self.list_scripts(options)
        if scripts:
            self.init_cluster_with_scripts(options, scripts)
        else:
            self.init_cluster_fallback(options)

    def init_cluster_with_scripts(self, options, scripts):
        cmd = ["psql", "-p", options.port, "-v", "ON_ERROR_STOP="]
        input_text = "\n".join(r"\i " + name for name in scripts)
        logging.debug("command with input: %r", cmd)
        if REAL:
            input = tempfile.TemporaryFile()
            input.write(input_text)
            input.seek(0)

            exec_command(cmd, stdin=input,
                         error_msg=("Error initialize cluster on port %r"
                                    % (options.port,)))

    def init_cluster_fallback(self, options):
        cmd = ["psql", "-p", options.port, "-v", "ON_ERROR_STOP=",
               "-f", "mpuls-init-cluster.sql"]
        logging.debug("command: %r", cmd)
        if REAL:
            exec_command(cmd)


class CreateDBCommand(DBCommand):
    """
    Creates a single new database
    
    To create a new single database use the following command::

	python mpulsdbadm.py create_db \\
		-p 5436 \\
		-d 20120823 \\
		--script-dir=/home/mpuls/wasku/db_config/db/install/
		--log-dir=/tmp \\

    Please note, that the command will create a ''result_list'' file in the
    directory of the mpulsdbadm command. Please make sure that the user invoking
    the command has the right to create new files in this directory.
    
    **Evaluation Database**
    
    If the new created database should be used as an evaluation server the database
    needs to be upgraded::

	python mpulsdbadm.py upgrade_db \\
		-p 5436 \\
		-d 20120823 \\
		--script-file=/home/mpuls/wasku/db_config/db/auswertung/evalserver.sql \\
		--log-dir=/tmp \\
    """
    
    name = "create_db"
    def add_options(self, parser):
        super(CreateDBCommand, self).add_options(parser)
        parser.add_option("--evalserver", action="store_true", dest="evalserver",
                          default=False)

    def check_options(self, options):
        valid = super(CreateDBCommand, self).check_options(options)
        if options.script_dir is None:
            for name in (self.name, "install"):
                if os.path.isdir(name):
                    options.script_dir = name
                    break
        return valid

    def execute_for_db(self, options, agency):
        full_db_name = self.full_db_name(agency)
        config_dir = os.path.join(options.script_dir, 'config')
        init_scripts = list_scripts(options.script_dir)
        config_scripts = list_scripts(config_dir)
        if not init_scripts:
            raise RuntimeError("No scripts found under %r" % options.script_dir)
        if not config_scripts:
            raise RuntimeError("No scripts found under %r" % config_dir)
        if options.evalserver:
            config_eval_dir = os.path.join(options.script_dir, 'config', 'evaluation')
            config_eval_scripts = list_scripts(config_eval_dir)
            if not config_eval_scripts:
                raise RuntimeError("No scripts found under %r" % config_eval_dir)
        else:
            config_eval_scripts = []

        self.create_db(options, agency, full_db_name)
        self.create_admin_users(options, agency, full_db_name)
        self.init_db(options, agency, full_db_name, init_scripts, config_scripts, config_eval_scripts)

    def create_db(self, options, agency, full_db_name):
        logging.info("Create database %r on port %r", agency, options.port)
        cmd = ["psql", "-t", "mpuls_maintenance", "-p", options.port, "-c",
                ("CREATE DATABASE %s WITH TEMPLATE = mpuls_template"
                 " ENCODING = 'UTF8';") % (full_db_name,)]
        logging.debug("command: %r", cmd)
        if REAL:
            exec_command(cmd, error_msg=("Error creating database %r on port %r"
                                         % (full_db_name, options.port)))

    def create_admin_users(self, options, agency, full_db_name):
        logging.debug("Entering save_password")
        cmd = ["psql", "-t", "mpuls_maintenance", "-p", options.port, "-c",
               ("SELECT mpuls_init_database('%s');" % (agency,))]
        logging.debug("command: %r", cmd)

        if REAL:
            logging.debug("Append password to result_list file")
            password = tempfile.TemporaryFile()
            exec_command(cmd, stdout=password,
                         error_msg=("Error calling mpuls_init_database"
                                    " for db %r on port %r."
                                    % (full_db_name, options.port)))

            password.seek(0)
            pw = password.readline()
            password.close()

            result = open("result_list", "a")
            result.write("%s\t%s" % (full_db_name, pw))
            result.close()
        else:
            logging.info("Try to append passwort to the result_list file")

    def init_db(self, options, agency, full_db_name, init_scripts, config_scripts, config_eval_scripts):
        cmd = ["psql", full_db_name, "-p", options.port]

        input_text="""\
         \set ON_ERROR_STOP
         SET ROLE ka_%(agency)s_adm;
         \set cm_group 'ka_%(agency)s_cm'
         \set admin_group 'ka_%(agency)s_admin'
         \set adm_ka_owner 'ka_%(agency)s_adm'
         \set db_name '%(dbname)s'
         \set fkz ''''%(agency)s'
         """ % dict(agency=agency, dbname=full_db_name)

        base_scripts = list_scripts('./sql/install/')
        scripts = []
        for script in base_scripts:
            if script == "./sql/install/06_master.sql":
                for iscript in init_scripts:
                    scripts.append(r"\i " + iscript)
            scripts.append(r"\i " + script)
        for cscript in config_scripts:
            scripts.append(r"\i " + cscript)

        if options.evalserver:
            eval_scripts = list_scripts('./sql/install/evaluation')
            for escript in eval_scripts:
                scripts.append(r"\i " + escript)
            for ecscript in config_eval_scripts:
                scripts.append(r"\i " + ecscript)
        
        input_text += "\n".join(scripts)

        logging.debug("command with input: %r", cmd)
        logging.debug("input text: %r", input_text)
        if REAL:
            input = tempfile.TemporaryFile()
            input.write(input_text)
            input.seek(0)

            exec_command(cmd, stdin=input,
                         error_msg=("Error initialize db %r on port %r"
                                    % (full_db_name, options.port)))


class UpgradeDBCommand(DBCommand):
    """
    Upgrades a database
    
    A database can be upgraded with the following command::

	python mpulsdbadm.py upgrade_db \\
		-p 5436 \\
		-d 20120823 \\
		--script-file=/home/mpuls/wasku/db_config/db/auswertung/evalserver.sql \\
		--log-dir=/tmp \\
    """
    name = "upgrade_db"

    def add_options(self, parser):
        super(UpgradeDBCommand, self).add_options(parser)
        parser.add_option("--old", action="store_true", default=False)
        parser.add_option("--trigger", action="store_true", default=False,
                          help="Disable last access trigger on upgrade")


    def execute_for_db(self, options, agency):
        full_db_name = self.full_db_name(agency)
        scripts = self.list_scripts(options)
        if not scripts:
            raise RuntimeError("No scripts found under %r" % options.script_dir)

        self.upgrade_db(options, agency, full_db_name, scripts)

    def upgrade_db(self, options, agency, full_db_name, scripts):
        cmd = ["psql", "-p", options.port, "-d", full_db_name]
        input_text = self.upgrade_build_input(options, agency, full_db_name,
                                              scripts)

        logging.debug("command with input: %r", cmd)
        logging.debug(input_text)

        if REAL:
            input = tempfile.TemporaryFile()
            input.write(input_text)
            input.seek(0)

            exec_command(cmd, stdin=input, stdout=self.logging_stream(),
                         stderr=self.logging_stream(),
                         error_msg=("Error upgrading db %r on port %r"
                                    % (full_db_name, options.port)))

    def upgrade_build_input(self, options, agency, full_db_name, scripts):
        old = options.old
        trigger = options.trigger
        input_text="""\
        \set cm_group 'ka_%(agency)s_cm'
        \set pb_group 'ka_%(agency)s_pb'
        \set admin_group 'ka_%(agency)s_admin'
        \set db_adm 'ka_%(agency)s_adm'
        \set adm_ka_owner 'ka_%(agency)s_adm'
        \set db_name 'ka_%(agency)s_db'
        \set fkz ''''%(agency)s'
        """ % dict(agency=agency)

        if not old:
            logging.debug("Using new upgrade scripts")
            input_text += "BEGIN;\n"

            if trigger:
                input_text += self.remove_trigger(options, agency, full_db_name)
            input_text += "SET ROLE :db_adm;\n"

        # Base sql scripts are located under './sql/install/'. Execute these scripts.

        for script in scripts:
            input_text += "\\i %s\n" % script

        if not old:
            for script in scripts:
                input_text += ("INSERT INTO db_history"
                               " (beschreibung, zeitpunkt)"
                               " VALUES ('%s', current_timestamp);\n"
                               % script)
            if trigger:
                input_text += "SET ROLE postgres;\n"
                input_text += self.add_trigger(options, agency, full_db_name)
            input_text += "COMMIT;\n"

        return input_text


    def remove_trigger(self, options, agency, full_db_name):
        return ("CREATE OR REPLACE FUNCTION"
                " set_lastaccess_time(masterid integer)\n"
                "RETURNS void AS ' ' LANGUAGE SQL;\n")

    def add_trigger(self, options, agency, full_db_name):
        return ("CREATE OR REPLACE FUNCTION"
                "  set_lastaccess_time(masterid integer)\n"
                "RETURNS void AS\n"
                "$$\n"
                "BEGIN\n"
                "   UPDATE ka_status_tbl_view set zugriff = now()\n"
                "   WHERE master_id = masterid;"
                "END;\n"
                "$$ LANGUAGE plpgsql VOLATILE;\n")



class DeleteDBCommand(DBCommand):
    """
    Deletes a database.
    
    To delete a database use the following command::

	python mpulsdbadm.py delete_db \\
		-p 5436 \\
		-d 20120823 \\
		--log-dir=/tmp
    """
    name = "delete_db"

    def execute_for_db(self, options, agency):
        full_db_name = self.full_db_name(agency)
        self.drop_db(options, agency, full_db_name)
        self.drop_db_users(options, agency, full_db_name)

    def drop_db(self, options, agency, full_db_name):
        cmd = ["dropdb", "-p", options.port, full_db_name]
        logging.debug("command: %r", cmd)
        if REAL:
            exec_command(cmd, error_msg=("Cannot drop db %s" % full_db_name))

    def drop_db_users(self, options, agency, full_db_name):
        cmd = ["psql", "-p", options.port, "-c", "SELECT rolname FROM pg_roles"]
        logging.debug("command: %r", cmd)

        if REAL:
            users = tempfile.TemporaryFile()
            exec_command(cmd, stdout=users,
                         error_msg="Finding users roles failed")
            r = re.compile('.*ka_%s_.*' % (agency,))
            users.seek(0)
            for l in users.readlines():
                if r.match(l):
                    self.drop_user(options, agency, full_db_name, l.strip())

    def drop_user(self, options, agency, full_db_name, username):
        cmd = ["dropuser", "-p", options.port, username]
        logging.debug("command: %r", cmd)
        return exec_command(cmd,
                            error_msg=("Error dropping user %s" % username))


def list_scripts(script_dir):
    scripts = []
    for name in os.listdir(script_dir):
        if re.match(r"[0-9][0-9].*\.sql$", name):
            scripts.append(os.path.join(script_dir, name))
    return sorted(scripts)


def single_argument_as_list(option, opt, value, parser):
    setattr(parser.values, option.dest, [value])

def file_contents_as_list(option, opt, value, parser):
    setattr(parser.values, option.dest, file_to_list(value))


def print_help():
    print >>sys.stderr, """\
    python dbscript.py init_cluster -p <port> [--stdout] [--dry-run]
    python dbscript.py create_db -d <db> | --dblist <dblistfile> -p <port>
            [--stdout] [--dry-run]
    python dbscript.py upgrade_db -d <db> | --dblist <dblistfile> -p <port>
             --upgradelist <upgradelistfile> | --upgradedir <upgrade-dir>
            [--stdout] [--old] [--dry-run]
    python dbscript.py delete_db -d <db> | --dblist <dblistfile> -p <port>
            [--stdout] [--dry-run]
"""

def file_to_list(fn1):
    f1 = open(fn1, "r")
    try:
        res = []

        for line in f1:
            line = line.strip()
            if line:
                if line.startswith("#"):
                    # ignore comment line
                    continue
                if "#" in line:
                    # remove comment from line
                    index = line.find("#")
                    line = line[:index].rstrip()
                res.append(line)

        return res
    finally:
        f1.close()


def exec_command(cmd_list, error_msg=None, stderr=None, stdout=None,
                 stdin=None):
    retcode = subprocess.call(cmd_list, stdout=stdout, stderr=stderr,
                              stdin=stdin)
    if retcode != 0:
        if error_msg:
            logging.error(error_msg)
        raise Exception(repr(cmd_list) + " failed")


def help_command_list(scriptname):
    print >>sys.stderr, "%s supports the following commands:" % (scriptname,)
    for name, cmd_class in list_commands():
        print >>sys.stderr, "    %s" % name


def main():
    if not len(sys.argv) > 1:
        help_command_list(os.path.basename(__file__))
        sys.exit(1)
    command_name = sys.argv[1]

    command_class = find_command(command_name)

    if command_class is None:
        print >>sys.stderr, "Unknown command: %r" % (command_name,)
        help_command_list(sys.argv[0])
        sys.exit(1)

    command = command_class()
    parser = command.get_option_parser()
    options, rest = parser.parse_args(sys.argv[1:])
    if not command.check_options(options):
        sys.exit(2)

    command.run(options)
    return 0


if __name__ == "__main__":
    main()
