#!/usr/bin/env python

import sys
from optparse import OptionParser
import psycopg2 as dbapi
import traceback
import calendar
import datetime

try:
    import sqlite3 as db
except ImportError:
    from pysqlite2 import dbapi2 as db

DEFAULT_PORT = 5432

START_DATE = '2011-09-01'
MONTH_RANGE = 16

SELECT_FKZ = "SELECT id, database FROM fkz_tbl"

ANON_TRANSFER = "SELECT wert = 'on' FROM ka_konfiguration_tbl WHERE bez = 'anon_transfer'"

SQL_ALL_CASES = """SELECT DISTINCT COUNT(id) FROM master_tbl_view m
WHERE (datum_ea_ende2 <= '%(edate)s' AND '%(sdate)s' <= coalesce(datum_cm_ende, now()) AND phase IN (2,3,4,5))"""

SQL_WASKA = """SELECT DISTINCT COUNT(id) FROM master_tbl
WHERE (datum_feststellung <= '%(edate)s' AND '%(sdate)s' <= coalesce(beendigung_5, now()) AND phase IN (2,3,4,5))  OR (datum_feststellung <= '%(edate)s' AND '%(sdate)s' <= coalesce(datum_ende_beratung, now()) AND phase IN (6,7))"""

SQL_WASKO = """SELECT DISTINCT COUNT(id) FROM master_tbl
WHERE (datum_cm_start <= '%(edate)s' AND '%(sdate)s' <= coalesce(datum_cm_ende, now()) AND phase IN (2,3,4,5))"""

SQL_ALL_CASES_DICT = {
        "waska3": SQL_ALL_CASES,
        "waska": SQL_WASKA,
        "wasko": SQL_WASKO
        }

LOAD_JOPPARTS = """\
SELECT wert FROM ka_konfiguration_tbl WHERE bez='job_percentage';
"""

UPDATE_TMPL_FKZ = """\
        UPDATE fkz_tbl SET job_percentage=? WHERE database=?;"""

def load_jobparts(cursor):
    return lookup(cursor, LOAD_JOPPARTS)

def load_active_cases(cursor, sdate, edate, app):
    if not sdate:
        sdate= "NOW()"
    if not edate:
        edate = "NOW()"
    query_tmpl = SQL_ALL_CASES_DICT.get(app, None)
    if not query_tmpl:
        query_tmpl = SQL_ALL_CASES
        print "Could not find app name in SQL_ALL_CASES_DICT, take default"
    replaces = {'sdate': sdate, 'edate': edate}
    query = query_tmpl % replaces
    return lookup(cursor, query)

def lookup(cursor, query):
    cursor.execute(query)
    row = cursor.fetchone()
    return row[0]

def calculate_job_percentage(cursor, start, monthrange, app):
    y,m,d = start.split('-')
    y = int(y)
    m = int(m)
    d = int(d)
    cases_list = []
    r = range(int(monthrange))
    for x in r:
        sd, ed = calendar.monthrange(y, m)
        s = "%s" % datetime.date(y,m,1)
        e = "%s" % datetime.date(y,m,ed)
        if datetime.date(y,m,1) > datetime.date.today():
            break
        cases = load_active_cases(cursor, s, e, app)
        cases_list.append((datetime.date(y,m,ed), cases))
        m+=1
        if m > 12:
            m=1
            y+=1
    count_used_month = 0
    cases_sum = 0
    for (date, cases) in cases_list:
        count_used_month += 1
        cases_sum += cases

    jobpart = load_jobparts(cursor)
    if jobpart == '' or jobpart =='0':
        return None
    else:
        return float(cases_sum) / (float(count_used_month) * float(jobpart))  

def usage(msg="", exit_code=0):
    print >> sys.stderr, """%s
usage:
%s -o <dump.db>
""" % (msg, sys.argv[0])
    sys.exit(exit_code)

def parse_port_list(option, opt, value, parser):
    setattr(parser.values, option.dest, [int(p) for p in value.split(',')])

def get_fkzs(cur):
    cur.execute(SELECT_FKZ)
    out = []
    while True:
        row = cur.fetchone()
        if not row:
            break
        out.append(row)
    return out

def handle(out_cur, fkz, port, startdate, month, app):
    fkz_id, database = fkz

    in_con, in_cur = None, None
    print >> sys.stderr, "processing (jp) '%s (%s)'... " % (database, port),

    try:
        in_con = dbapi.connect(database=database, port=port)
        in_cur = in_con.cursor()

        in_cur.execute(ANON_TRANSFER)
        row = in_cur.fetchone()
        if not row or not row[0]:
            print >> sys.stderr, "No transfer allowed by '%s'" % database
            return
        jobpercentage = calculate_job_percentage(in_cur,startdate, int(month), 
                                                 app)
        if not jobpercentage is None:
            out_cur.execute(UPDATE_TMPL_FKZ,(jobpercentage, database))
    finally:
        if in_cur:
            in_cur.close()
        if in_con:
            in_con.close()


def main():
    parser = OptionParser()
    parser.set_defaults(port=[DEFAULT_PORT], application='waska3')
    parser.add_option("-o", "--out",
                      help="Filename of the output database")
    parser.add_option("-p", "--port", action="callback",
                      callback=parse_port_list, type="string",
                      help=("Comma-separated port numbers of the PostgreSQL"
                            " databases"))
    parser.add_option("-s","--startdate",
                      help="ISO start date of the job percentage calc")
    parser.add_option("-m","--months",
                      help=("Number of months max used for job percentage"
                            " calc"))
    parser.add_option("-w", "--application",
                      help="Name of the app for mapping to where statement")

    opts,args = parser.parse_args()
    if opts.out is None:
        usage("Error: missing output database parameter -o", 1)
    if opts.startdate is None:
        usage("Error: missing startdate parameter -s", 1)
    if opts.months is None:
        usage("Error: missing months parameter -m", 1)

    out_con, out_cur = None, None

    try:
        out_con = db.connect(opts.out)
        out_cur = out_con.cursor()

        try:
            fkzs = get_fkzs(out_cur)
            for p in opts.port:
                for fkz in fkzs:
                    try:
                        sd = opts.startdate
                        m = opts.months
                        a = opts.application
                        handle(out_cur, fkz, p, sd, m, a) 
                        print >> sys.stderr, "OK"
                    except:
                        print >> sys.stderr, "Failed"
                        traceback.print_exc(file=sys.stderr)

            out_con.commit()
        except:
            out_con.rollback()
            raise
    finally:
        if out_cur:
            out_cur.close()
        if out_con:
            out_con.close()

if __name__ == '__main__':
    main()
