# -*- coding: utf-8 -*-
# Copyright 2009 Intevation GmbH, Germany, <info@intevation.de>
# Authors:
# Sascha L. Teichmann <teichmann@intevation.de>
#


class DBException(Exception):

    pass


class DB(object):

    """Base class for DB interfaces for libformed.

    This class defines the basic DB interface and supported for nested
    transactions.  It cannot be used directly to get access to a
    database.  Derived classes need to override at least the methodes
    acquire_cursor, release_cursor, do_rollback and do_commit and they
    need to provide a means to get a database connection.
    """

    def __init__(self):
        self.transaction_level = 0
        self.rolled_back = False

    def acquire_cursor(self):
        """Return a cursor.
        The cursor must be release with a call to release_cursor()
        """
        raise NotImplementedError

    def release_cursor(self, cur):
        """Release a cursor acquired with acquire_cursor()."""
        raise NotImplementedError

    def begin(self):
        """Start a transaction.

        It is safe to call this method even when a transaction has
        already been started on this interface instance.  Such
        additional calls start a nested transaction.  Each call to
        begin() must be matched by either a call to commit() or
        rollback().  Once rollback has been called, further nested calls
        to begin raise a DBException.
        """
        if self.rolled_back:
            raise DBException("Trying to begin new (sub-)transaction while"
                              " surrounding transaction has errors")
        self.transaction_level += 1

    def rollback(self):
        """Rollback the current top-level transaction.

        In all cases this method rolls back the current database
        transaction, regardless of the nesting level.
        """
        if not self.rolled_back:
            self.rolled_back = True
            self.do_rollback()
        self.transaction_level -= 1

    def do_rollback(self):
        """Perform the actual database rollback.  Override in derived classes.
        """
        raise NotImplementedError

    def commit(self):
        """Commit the current transaction.
        If called for a nested transaction, nothing really happens apart
        from the book-keeping required for nested transactions.  If
        called for the top-level transaction, the transaction will be
        committed unless the transaction was actually rolled back
        earlier.
        """
        self.transaction_level -= 1
        if self.transaction_level == 0:
            if not self.rolled_back:
                self.do_commit()

    def do_commit(self):
        """Perform the actual database commit.  Override in derived classes."""
        raise NotImplementedError


# vim:set ts=4 sw=4 si et sta sts=4 enc=utf-8 fenc=utf-8 :
