# -*- coding: utf-8 -*-
# Copyright 2010 Intevation GmbH, Germany, <info@intevation.de>
# Authors:
# Bernhard Herzog <bernhard.herzog@intevation.de>

"""Structure instance trees that match the meta structure trees."""

import uuid

from formed.meta.data import RepeatNode
from formed.meta.exprtree import VariableProvider


class StructureInstanceNode(object):

    """Node of the structure instance tree.

    Public read-only instance variables:

       uuid -- The uuid of the node

       id -- The database id of the node.  None if the node is not in
             the database

       meta_node -- The corresponding meta structure node.
    """

    def __init__(self, meta_node, id, uuid=None, leaves=None,
                 repeat_groups=None):
        """Initialize the instance.
        The parameters meta_node, id and uuid are simply bound to the
        corresponding instance variables.  The parameters leaves and
        repeat_groups, if given and not None, should be dictionaries.

        The leaves dictionary maps the names of the leaves of this
        instance to the values of the leaves.

        The repeat_groups dictionary may have one of two structures:

         - a dictionary mapping the names of the repeat groups to
           dictionaries mapping the instance ids to the actual repeat
           group instances.

         - a dictionary mapping the names of the repeat groups to lists
           with the repeat group instances. In this case the lists will
           be replaced with dictionaries mapping the repeat group ids to
           the instances with new instance ids assigned to all the
           instances.

        Mixing these two structures in one repeat groups dictionary is
        not allowed. If its accepted without errors, it's an
        implementation detail that the caller should not rely on.

        The names of the leaves and repeat groups must be the ones
        defined in the meta_node.  All leaves must occur in the leaves
        dictionary.

        If the leaves dictionary is not given or None, a dictionary with
        default values is created.  The default values are taken from
        the formed description.
        """
        self.meta_node = meta_node
        self.id = id
        self.uuid = uuid
        if leaves is None:
            leaves = self.default_leaves()
        self.leaves = leaves
        if repeat_groups is None:
            repeat_groups = self.default_repeat_groups()
        self.repeat_groups = self.convert_repeat_groups_dict(repeat_groups)

    def convert_repeat_groups_dict(self, repeat_groups):
        converted = dict()
        for name, instances in repeat_groups.items():
            if not isinstance(instances, dict):
                instances = self.convert_repeat_group_instance_list(instances)
            converted[name] = instances
        return converted

    def convert_repeat_group_instance_list(self, instances):
        converted = dict()
        for node in instances:
            node.id = self.generate_rg_id(converted)
            converted[node.id] = node
        return converted

    def default_leaves(self):
        leaves = dict()
        for leaf in self.meta_node.leaves.values():
            leaves[leaf.getName()] = leaf.findDefault()
        return leaves

    def default_repeat_groups(self):
        repeat_groups = dict()
        for rg_meta_node in self.meta_node.repeat_groups.itervalues():
            instances = repeat_groups[rg_meta_node.name] = list()
            for i in range(rg_meta_node.meta_node.getMin()):
                instances.append(self.__class__(rg_meta_node, None))
        return repeat_groups

    def generate_rg_id(self, instances):
        """Return a repeat group id not yet used in instances."""
        return max([0] + instances.keys()) + 1

    def get(self, name):
        local_name = name[0]
        restname = name[1:]

        if isinstance(local_name, tuple):
            item = self.get_repeat_group(*local_name)
            if restname:
                item = item.get(restname)
        else:
            if local_name in self.meta_node.leaves:
                item = self.get_leaf(local_name)
            elif local_name in self.meta_node.repeat_groups:
                item = self.get_repeat_groups(local_name)
            else:
                raise ValueError("Name %r is not known as leaf or repeat group"
                                 % local_name)
        return item

    def set_value(self, name, value):
        local_name = name[0]
        restname = name[1:]

        if isinstance(local_name, tuple):
            item = self.get_repeat_group(*local_name)
            if restname:
                item.set_value(restname, value)
        else:
            meta_leaf = self.meta_node.leaves.get(local_name)
            if meta_leaf is None:
                raise ValueError("Name %r is not known as a leaf" % local_name)
            self.set_leaf(local_name, value)

    def get_leaf(self, meta_name):
        return self.leaves[meta_name]

    def set_leaf(self, meta_name, value):
        self.leaves[meta_name] = value

    def get_repeat_groups(self, rg_name):
        if rg_name not in self.meta_node.repeat_groups:
            raise ValueError("Node %r has no repeat group child %r"
                             % (self.meta_node.name, rg_name))
        return self.repeat_groups.setdefault(rg_name, dict())

    def get_repeat_group(self, rg_name, rg_id):
        rg = self.get_repeat_groups(rg_name).get(rg_id)
        if not rg:
            raise ValueError("Unkown RG ID %r for repeat group %r"
                             % (rg_id, rg_name))
        return rg

    def create_repeat_group(self, rg_name):
        node = self.__class__(self.meta_node.repeat_groups[rg_name], None,
                              uuid=uuid.uuid4())
        instances = self.get_repeat_groups(rg_name)
        node.id = self.generate_rg_id(instances)
        instances[node.id] = node
        return node

    def delete_repeat_group(self, rg_name, rg_id):
        rgs = self.get_repeat_groups(rg_name)
        if rg_id in rgs:
            del rgs[rg_id]

    def make_unique(self):
        """Assign new UUIDs to the node and contained repeat groups recursively
        """
        self.uuid = unicode(uuid.uuid4())
        for rgs in self.repeat_groups.itervalues():
            for rg in rgs.itervalues():
                rg.make_unique()

    def to_primitive(self):
        """Return the instance as a dictionary that can be pickled.
        Use from_primitive to create a new instance from such a dictionary.
        """
        return dict(name=self.meta_node.name, id=self.id, uuid=self.uuid,
                    leaves=self.leaves.copy(),
                    rgs=[(name, [rg.to_primitive() for rg in rgs.values()])
                         for name, rgs in self.repeat_groups.items()])

    @classmethod
    def from_primitive(cls, meta_node, primitive):
        """Create a new instance for meta_node reconstructed from primitive.
        The parameter primitive should be a dictionary previously
        returned by the to_primitive method.
        """
        if primitive["name"] != meta_node.name:
            raise ValueError("Name from primitive dictionary (%r)"
                             " does not match meta node (%r)"
                             % (primitive["name"], meta_node.name))

        repeat_groups = dict()
        for name, instances in primitive["rgs"]:
            meta = meta_node.repeat_groups[name]
            repeat_groups[name] = dict((i["id"], cls.from_primitive(meta, i))
                                       for i in instances)
        return cls(meta_node, primitive["id"],
                   uuid=primitive["uuid"],
                   leaves=primitive["leaves"].copy(),
                   repeat_groups=repeat_groups)


def pprint_structure_instance_node(node, indent=""):
    """Recursively pretty-print a structure instance node to stdout"""
    print ("%s%s:%s:%s: %d leaves, %d repeat groups"
           % (indent, node.meta_node.name, node.id, node.uuid,
              len(node.leaves), len(node.repeat_groups)))
    indent = "    " + indent
    for key, item in sorted(node.leaves.items()):
        print "%s%s: %r" % (indent, key, item)
    for name, rgs in sorted(node.repeat_groups.items()):
        for key, item in sorted(rgs.items()):
            pprint_structure_instance_node(item, indent)


class StructureInstanceTree(object):

    """Structure instance tree.

    Public read-only instance variables:

      root -- The root structure instance node

      meta_structure -- The corresponding structure meta tree

      uuid -- The uuid of the tree.  Identical to the uuid of the root
              node.

      id -- The database id of the instance.  Identical to the id of the
            root node.
    """

    def __init__(self, meta_structure, root=None, id=None, uuid=None):
        self.meta_structure = meta_structure
        if root is not None:
            self.root = root
        else:
            self.root = self.instantiate_root(meta_structure.root, id, uuid)

    @property
    def uuid(self):
        return self.root.uuid

    @property
    def id(self):
        return self.root.id

    def instantiate_root(self, meta_root, id, uuid):
        return StructureInstanceNode(meta_root, id, uuid=uuid)

    def get(self, name):
        """Return the value of the item indicated by name.
        The argument name can be of two forms: it may be a string using
        the legacy syntax used by formed.instance.tree, or a new item
        specification.

        If the name is a string containing at least one colon (':'), it
        is assumed to be a legacy name.  Legacy names are converted to
        item specifications and then treated like those.

        If the name is not a legacy name, it's an item specification.
        Item specifications have form (rgid1, rgid2, ..., itemname).
        The last part of that tuple is the name of the item as used in
        the meta tree.  The rgidN values are integers specifying the ids
        of the repeat groups that contain the instance of the item.  The
        meta structure defines how many repeat groups there are in the
        hierarchy above the item, so the number of repeat group ids
        required is always known.

        If itemname ist the name of a repeat group, there are two cases
        to consider, though.  The item specification can refer to all
        instances of the repeat group or to a particular instance.  If
        the the number of repeat group ids given in the item
        specification is equal to the number of the repeat groups above
        the item named by itemname, the specification refers to all
        instances.  If there's one more repeat group id, it is
        interpreted as the id of a particular instance of the repeat
        group named by itemname.

        Another special case is items that are not contained in repeat
        groups.  In this case the item name can be given directly as a
        string.  It does not have to be wrapped in a single-element
        tuple.
        """
        return self.root.get(self.full_name(name))

    def set_value(self, name, value):
        self.root.set_value(self.full_name(name), value)

    def create_repeat_group(self, name):
        """Create a new instance of the repeat group indicated by name.
        The parameter name should not refer to a specific existing
        instance of the repeat group, i.e. it should have exactly as
        many ids as there are repeat group parents as described in the
        documentation of the get method.
        """
        parent, relative_name = self._get_repeat_group_parent(name)
        parent.create_repeat_group(relative_name)

    def delete_repeat_group(self, name):
        """Delete the repeat group instance indicated by name.
        The parameter name should refer to a specific existing instance
        of the repeat group, as described in the documentation of the
        get method.
        """
        parent, relative_name = self._get_repeat_group_parent(name)
        parent.delete_repeat_group(*relative_name)

    def _get_repeat_group_parent(self, name):
        full_name = self.full_name(name)
        if len(full_name) == 1:
            parent = self.root
        else:
            parent = self.root.get(full_name[:-1])
        return parent, full_name[-1]

    def full_name(self, name):
        # name is either simply a string with a field name, or a tuple
        # (rgid1, ..., rgidN, fieldname)
        if isinstance(name, basestring):
            if ":" in name:
                name = self.parse_legacy_name(name)
            else:
                name = (name,)

        path = self.meta_structure.structure_path(name[-1])
        num_ids = len(name) - 1
        path_len = len(path)
        if isinstance(path[-1], RepeatNode):
            if num_ids not in (path_len, path_len - 1):
                raise ValueError("The full name for repeat group %r must have"
                                 " either %d or %d ids."
                                 % (name, path_len - 1, path_len))
        else:
            if num_ids != path_len - 1:
                raise ValueError("The full name for item %r must have exactly "
                                 " %d ids." % (name, path_len - 1))
        full_name = []
        for i in range(num_ids):
            full_name.append((path[i].getName(), name[i]))
        if num_ids < path_len:
            full_name.append(name[-1])
        return full_name

    def parse_legacy_name(self, legacy_name):
        """Convert a legacy instance item name into the new item specification.
        The legacy name is a string consisting of colon-separated
        fields.  The first field is the item's meta name.  The rest of
        the fields are the ids of the repeat groups containing the item,
        starting with the one immediately containing the instance item,
        followed by the id of the repeat group containing that repeat
        group, and so on.  The master table itself is treated as a
        repeat group in this context, so the last repeat group id is
        always the case id.
        """
        assert ":" in legacy_name
        parts = legacy_name.split(":")
        parts.reverse()
        for i in range(len(parts) - 1):
            parts[i] = int(parts[i])
        if parts[0] != self.root.id:
            raise ValueError("Legacy name must contain the case id")
        return tuple(parts[1:])

    def get_all(self, name):
        """Return all values for the leaf named by name.
        The return value is a sequence of tuples of the form (NAME,
        VALUE) where NAME is a name usable for the get method to
        retrieve the value and VALUE is the current value of the item.
        """
        def generate_items(node, path):
            child_meta = path[0]
            child_name = child_meta.getName()
            if isinstance(child_meta, RepeatNode):
                for rg in node.get_repeat_groups(child_name).values():
                    for partial_id, value in generate_items(rg, path[1:]):
                        yield ((rg.id,) + partial_id, value)
            else:
                yield ((child_name,), node.get((child_name,)))

        path = self.meta_structure.structure_path(name)

        return list(generate_items(self.root, path))

    def iter_all_items(self):
        """Yield all values in the tree as (key, value) pairs."""
        nodes = [self.meta_structure.root]
        while nodes:
            node = nodes.pop(0)
            for name in node.leaves:
                for item in self.get_all(name):
                    yield item
            nodes.extend(node.repeat_groups.values())

    def make_unique(self):
        """Assign new UUIDs to all nodes recursively."""
        self.root.make_unique()

    def to_primitive(self):
        """Return the tree as a dictionary that can be pickled.
        Use from_primitive to create a new instance from such a dictionary.
        """
        return dict(formed_version=self.meta_structure.get_version(),
                    tree=self.root.to_primitive())

    @classmethod
    def from_primitive(cls, meta_structure, primitive):
        """Create a new tree for meta_structure reconstructed from primitive.
        The parameter primitive should be a dictionary previously
        returned by the to_primitive method.
        """
        expected_version = meta_structure.get_version()
        found_version = primitive["formed_version"]
        if expected_version != found_version:
            raise ValueError("primitive instance tree has wrong version."
                             " Expected %r, found %r."
                             % (expected_version, found_version))

        return cls(meta_structure,
                   StructureInstanceNode.from_primitive(meta_structure.root,
                                                        primitive["tree"]))


class StructureVariableProvider(VariableProvider):

    def __init__(self, tree, reference, shadow=None):
        self.tree = tree

        if reference is None:
            self.ref_ids = ()
            self.ref_path = ()
        else:
            self.ref_ids, self.ref_path = self._ref_paths(reference)

        if shadow is None:
            shadow = dict()
        self.shadow = shadow

    def _ref_paths(self, reference):
        ids = reference[:-1]
        path = self.tree.meta_structure.structure_path(reference[-1])[:len(ids)]
        return (ids, path)

    def _full_key(self, name, reference):
        path = self.tree.meta_structure.structure_path(name)[:-1]
        required_ids = len(path)

        # If name is not contained in any repeat groups the key is just
        # the name as a singleton tuple
        if not required_ids:
            return (name,)

        # At least one RG id is required, so we have to look up the name
        # relative to the reference.
        if reference is None:
            ref_ids = self.ref_ids
            ref_path = self.ref_path
        else:
            ref_ids, ref_path = self._ref_paths(reference)

        # The name must be a child of an ancestor of the reference, i.e.
        # of one of the RGs in ref_path. Otherwise there's no unique key
        # because it't not clear which RG instance to use.
        #
        # We need exactly required_ids ids resp. ancestors. So all we
        # need to do is check whether the last element of path is
        # element (required_ids - 1) of ref_path if that exists.
        # Otherwise we cannot look up the value for name and thus cannot
        # return the key for it.
        if required_ids > len(ref_path):
            raise ValueError("Cannot determine key for %r."
                             " Requires %d ids, more than path %r has elements."
                             % (name, required_ids, zip(ref_ids, ref_path)))

        if path[-1] is ref_path[required_ids - 1]:
            return ref_ids[:required_ids] + (name,)

        raise ValueError("Cannot determine key for %r."
                         " Path entry for %r not found in path %r."
                         % (name, path[-1], zip(ref_ids, ref_path)))

    def getValueForVar(self, name, reference=None):
        key = self._full_key(name, reference)
        if key in self.shadow:
            return (key, self.shadow[key])
        return (key, self.tree.get(key))

    def getAllValuesForVar(self, name):
        return [(key, self.shadow.get(key, value))
                for key, value in self.tree.get_all(name)]

    def get_repeat_group_instances(self, name):
        key = self._full_key(name, None)
        key_ids = key[:-1]
        return [key_ids + (i, name) for i in self.tree.get(key).keys()]

