# -*- coding: utf-8 -*-
# Copyright 2010 Intevation GmbH, Germany, <info@intevation.de>
# Authors:
# Bernhard Herzog <bernhard.herzog@intevation.de>

"""Meta structure tree.

The meta structure tree is a simplified version of the normal FormEd
meta tree.  The structure tree only contains the repeat group structure
and the leaves that describe the actual fields.
"""

from formed.meta.data import WidgetCollector, RepeatNode, VISIT_CONTINUE


class StructureWidgetCollector(WidgetCollector):

    def handleRepeatNode(self, repeat):
        # visit all repeats
        self.widgets.append(repeat)
        return VISIT_CONTINUE


def structure_path(nc):
    """Return a list with the RepeatNode ancestors of nc from the root down.
    The last item in the list is the NodeComponent nc itself.
    """
    return [item for item in nc.pathToRoot()
                 if isinstance(item, RepeatNode) or item is nc]


class StructureNode(object):

    """
    One node in the structure tree.

    Public read-only instance variables:

      name -- name of the node.  For repeat groups, it's the repeat
              group name.  For the root node, the name is None.

      meta_node -- RepeatNode instance from the formed meta tree.  For
                   the root node, the meta_node is None

      leaves -- Dictionary mapping leaf names to corresponding Leaf
                instance

      repeat_groups -- Dictionary mapping repeat group names to the
                       StructureNode instance for the repeat group.
    """

    def __init__(self, meta_node, ordered_leaves, ordered_repeat_groups):
        """Instantiate the structure meta node.  Parameters:

          meta_node -- the formed meta node for the repeat group, None
              for the root node.

          ordered_leaves -- A list of the formed meta nodes of the
              leaves contained in the repeatgroup in the order in which
              they occur in the formed meta tree.  This order is
              important for the XML import because the schema requires
              the formed order for the leaves.

          ordered_repeat_groups -- A list of the StructureNode instances
              for the repeat groups contained in the new node.  Like for
              ordered_leaves, the oder should be the one from the formed
              meta tree.
        """
        self.meta_node = meta_node
        if meta_node is not None:
            name = meta_node.getName()
        else:
            name = None
        self.name = name
        self.ordered_leaves = ordered_leaves
        self.leaves = dict((widget.getName(), widget)
                           for widget in ordered_leaves)
        self.ordered_repeat_groups = ordered_repeat_groups
        self.repeat_groups = dict((rg.name, rg)
                                  for rg in ordered_repeat_groups)

    def __repr__(self):
        return "<StructureNode %x name=%r>" % (id(self), self.name)

    def is_root(self):
        """Return whether the node is the root node."""
        return self.name is None


def pprint_structure_node(node, indent="", print_leaves=True):
    """Recursively pretty-print the structure node to stdout."""
    name = node.name
    if name is None:
        name = "master"
    print ("%s%s: %d leaves, %d repeat groups"
           % (indent, name, len(node.leaves), len(node.repeat_groups)))
    indent = "    " + indent
    if print_leaves:
        for key, item in sorted(node.leaves.items()):
            print "%s%s: %s" % (indent, key, item)
    for key, item in sorted(node.repeat_groups.items()):
        pprint_structure_node(item, indent, print_leaves=print_leaves)


class StructureTree(object):

    """The structure tree.

    Public read-only instance variables:

      meta_tree -- The Document instance with the formed meta tree.

      root -- The root StructureNode instance
    """

    def __init__(self, meta_tree):
        """Initialize the structure tree from the formed meta tree meta_tree."""
        self.meta_tree = meta_tree
        self.path_cache = dict()
        self.build_structure_tree()

    def get_version(self):
        return self.meta_tree.getRoot().getVersion()

    def structure_path(self, name):
        return self.path_cache[name]

    def build_structure_tree(self):
        raw_tree, order = self.build_raw_structure_tree()
        self.root = self.convert_raw_tree_dict(None, raw_tree, order)

    def build_raw_structure_tree(self):
        order = []
        raw_tree = dict()

        collector = StructureWidgetCollector()
        self.meta_tree.visit(collector.visitor)

        for widget in collector.widgets:
            mapping = raw_tree
            path = structure_path(widget)
            if widget.getName() in self.path_cache:
                raise ValueError("Widget name %r used multiple times"
                                 % (widget.getName(),))
            self.path_cache[widget.getName()] = path

            order.append(widget.getName())

            if isinstance(widget, RepeatNode):
                continue

            for nc in path:
                name = nc.getName()
                if not name:
                    raise ValueError("Formed node component without name: %r"
                                     % (nc,))
                if nc is widget:
                    mapping[name] = nc
                else:
                    mapping = mapping.setdefault(name, dict())

        return raw_tree, order

    def convert_raw_tree_dict(self, name, raw_dict, full_order):
        leaves = []
        repeat_groups = []
        for child_name in full_order:
            child = raw_dict.pop(child_name, None)
            if child is None:
                continue
            if isinstance(child, dict):
                repeat_groups.append(self.convert_raw_tree_dict(child_name,
                                                                child,
                                                                full_order))
            else:
                leaves.append(child)

        # at this point all items from raw_dict should have been
        # converted
        assert not raw_dict

        if name:
            root_node = self.meta_tree.findByClassAndName(RepeatNode, name)
        else:
            root_node = None

        return StructureNode(root_node, leaves, repeat_groups)
